return {
  
  features  = {
    kern    = true,
    liga    = true,
    trep    = true,
    tlig    = true,
    script  = "latn",
    lang    = "dflt",
    --stretch = 30,
    --shrink  = 20,
    --step    = 10,
  },

  normal = {
    "Normal", "Regular", "Book",
  },
  
}
